/*
 * Copyright (c) 2005
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* This is an example implementation of the base64 encoding as specified
 * in RFC 2045 (Multipurpose Internet Mail Extensions - Part One), see
 * http://www.ietf.org/rfc/rfc2045.txt for more information.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define BASE64_CORE
#include "base64.h"

/* base64 table for encoding */
static char base64_alphabet[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/* inverse table for fast lookup when decoding */
static int base64_inverse[ASCII_MAX];

/* the inverse table only needs once to be built */
static void base64_build_inverse(void)
{
	static int i;

	/* note that i contains ALPH_LEN after the first run, so this is
	 * an indicator that this initialization has already been done
	 */
	if (!i)
	{
		for (i = 0; i < ALPH_LEN; i++)
			base64_inverse[(int)base64_alphabet[i]] = i;
	}
}

/* decide whether a given character is a valid index to the alphabet,
 * therefore excluding the padding character '='
 */
static int valid_input(int c)
{
	if (c >= 'A' && c <= 'Z')
		return 1;

	if (c >= 'a' && c <= 'z')
		return 1;

	if (c >= '0' && c <= '9')
		return 1;

	if (c == '+' || c == '/')
		return 1;

	return 0;
}

/* This function takes a pointer to a piece of memory that holds binary
 * data to be encoded. The output buffer MUST be large enough to take
 * the encoded data, no checking will be done. The user MUST take care
 * of this constrain.
 */
int base64_encode(const void *in, size_t len, void *out)
{
	unsigned char *in_c = (unsigned char*) in;
	unsigned char *out_c = (unsigned char*) out;
	int input_pos, i;
	unsigned char input_buffer[DECODED_BUFFER_SIZE];
	int output_indices[ENCODED_BUFFER_SIZE];
	int out_len = 0;

	/* the input_pos counter remembers the position in the 24-bit
	 * sequence -- if it is DECODED_BUFFER_SIZE, 24 bits are together
	 * and can be encoded. If the end of input is encountered before
	 * that happens, the input_pos 'knows' how many zero-bits habe
	 * to be padded after the loop.
	 */
	input_pos = 0;
	while (len)
	{
		/* take the next binary character from input */
		input_buffer[input_pos++] = *in_c++;
		len--;

		if (input_pos != DECODED_BUFFER_SIZE)
			continue;

		input_pos = 0;

		/* note that all integers are unsigned, would be probably
		 * funny otherwise...
		 */
		output_indices[0] = (input_buffer[0] & 0xfc) >> 2;
		output_indices[1] = ((input_buffer[0] & 0x03) << 4) |
			((input_buffer[1] & 0xf0) >> 4);
		output_indices[2] = ((input_buffer[1] & 0x0f) << 2) |
			((input_buffer[2] & 0xc0) >> 6);
		output_indices[3] = input_buffer[2] & 0x3f;

		/* not worth looping... */
		*out_c++ = base64_alphabet[output_indices[0]];
		*out_c++ = base64_alphabet[output_indices[1]];
		*out_c++ = base64_alphabet[output_indices[2]];
		*out_c++ = base64_alphabet[output_indices[3]];

		/* count the bytes produced as output */
		out_len += 4;
	}

	/* If the input wasn't an integral number of 24 bits, some
	 * zero-stuffing and padding will be done here.
	 *
	 * There are three scenarios:
	 * 1) no padding is needed, input_pos is zero, code will be skipped
	 * 2) 1 byte was read, is in base64 2 bytes, add 2 bytes padding
	 * 3) 2 bytes were read, is in base64 3 bytes, add 1 byte padding
	 */
	if (input_pos)
	{
		/* fill up with zeros */
		for (i = input_pos; i < DECODED_BUFFER_SIZE; i++)
			input_buffer[i] = 0;

		/* same encoding as above... */
		output_indices[0] = (input_buffer[0] & 0xfc) >> 2;
		output_indices[1] = ((input_buffer[0] & 0x03) << 4) |
			((input_buffer[1] & 0xf0) >> 4);
		output_indices[2] = ((input_buffer[1] & 0x0f) << 2) |
			((input_buffer[2] & 0xc0) >> 6);
		output_indices[3] = input_buffer[2] & 0x3f;

		/* ...but only the valid characters are put into the output,
		 * the rest is padded with the PADDING_CHAR which is '='.
		 */
		for (i = 0; i < ENCODED_BUFFER_SIZE; i++)
		{
			if (i <= input_pos)
				*out_c++ = base64_alphabet[output_indices[i]];
			else
				*out_c++ = PADDING_CHAR;
			out_len++;
		}
	}

	/* return the number of bytes in the output buffer */
	return out_len;
}

/* This function takes a buffer with base64-encoded data an decodes
 * it into the output buffer, which MUST be large enough to take
 * the data. This is NOT checked!
 */
int base64_decode(const void *in, size_t len, void *out)
{
	unsigned char *in_c = (unsigned char*) in;
	unsigned char *out_c = (unsigned char*) out;
	int i, input_pos;
	int input_buffer[ENCODED_BUFFER_SIZE];
	unsigned char output_buffer[DECODED_BUFFER_SIZE];
	int out_len = 0;

	/* build the reverser table, if necessary */
	base64_build_inverse();

	/* as in base64_encode(), this pointer remembers the current
	 * position in the 24-bit-sequence, which happens to be a 32-
	 * bit-sequence in encoded form, ENCODED_BUFFER_SIZE is 4.
	 */
	input_pos = 0;

	/* if we meet a PADDING_CHAR, real data is finished */
	while (len && *in_c != PADDING_CHAR)
	{
		/* skip whitespaces, makes it easier to decode whole base64 encoded
		 * files, which have LF or CRLF at 72 characters
		 */
		if (isspace(*in_c))
		{
			in_c++;
			len--;
			continue;
		}

		/* if an invalid character is encountered the user is told */
		if (!valid_input(*in_c))
			return -1;

		input_buffer[input_pos++] = *in_c & 0x7f;	/* paranoia */

		in_c++;
		len--;

		/* loop until 4 x 6 significant bits are together */
		if (input_pos != ENCODED_BUFFER_SIZE)
			continue;

		/* reset the counter, of course */
		input_pos = 0;

		/* unsigned bit arithemetic, do not try this with signed */
		output_buffer[0] = (base64_inverse[input_buffer[0]] << 2) |
			(base64_inverse[input_buffer[1]] >> 4);
		output_buffer[1] = (base64_inverse[input_buffer[1]] << 4) |
			(base64_inverse[input_buffer[2]] >> 2);
		output_buffer[2] = (base64_inverse[input_buffer[2]] << 6) |
			base64_inverse[input_buffer[3]];

		/* not worth looping */
		*out_c++ = output_buffer[0];
		*out_c++ = output_buffer[1];
		*out_c++ = output_buffer[2];

		/* counter for bytes in output buffer */
		out_len += 3;
	}

	/* if the input was padded, because the original input wasn't
	 * an integral number of 24 bits, this will be corrected
	 *
	 * There are three scenarios:
	 * 1) no padding was done, input pos is zero, code will be skipped
	 * 2) 2 bytes padding, input_pos is 2 (because 2 bytes were valid)
	 * 3) 1 byte padding, input_pos is 3 (because 3 bytes wer valid)
	 */
	if (input_pos)
	{
		/* decode whatever is in the input buffer, care of not
		 * putting illegal data into the output is taken below
		 */
		output_buffer[0] = (base64_inverse[input_buffer[0]] << 2) |
			(base64_inverse[input_buffer[1]] >> 4);
		output_buffer[1] = (base64_inverse[input_buffer[1]] << 4) |
			(base64_inverse[input_buffer[2]] >> 2);

		/* output either one or two bytes, depending on input_pos */
		for (i = 0; i < input_pos - 1; i++)
		{
			*out_c++ = output_buffer[i];
			out_len++;
		}
	}

	/* return the number of bytes in the output buffer */
	return out_len;
}

